__description__ = "a module that houses routines to parse our configuration object"
__author__ = "Reed Essick (reed.essick@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import os
from numpy import inf
from collections import defaultdict

from six.moves import configparser

### non-standard libraries
from idq import synthetic

#-------------------------------------------------

def path2config(path, verbose=False):
    """
    read in the config from a path
    """
    if verbose:
        print('reading config from: '+path)
    if not os.path.exists(path):
        raise OSError('file does not exist!: '+path)
    config = configparser.SafeConfigParser()
    config.read(path)
    return config

def config2bounds(bounds):
    """
    a special parser for setting up bounds for trigger sets
    """
    ans = {}
    for bound in bounds.strip().split('\n'):
        key, min_val, max_val = bound.strip().split()
        try:
            min_val = int(min_val)
        except ValueError:
            min_val = float(min_val)
        try:
            max_val = int(max_val)
        except ValueError:
            max_val = float(max_val)

        ans[key] = (min_val, max_val)
    return ans

def config2param_dist(hyperparams):
    """
    a special parser for setting up randomized cross-validation bounds, distribution
    """
    ans = {}
    for hyperparam in hyperparams.strip().split('\n'):
        try:
            key, min_val, max_val, dist, dist_type = hyperparam.strip().split()
            ans[key] = (float(min_val), float(max_val), dist, dist_type)
        except: ### NOTE: may be fragile
            key, list_ = hyperparam.strip().split(None, 1)
            ans[key] = eval(list_)
    return ans

def config2param_grid(hyperparams):
    """
    a special parser for setting up grid-based cross-validation bounds, distribution, num_samples
    """
    ans = {}
    for hyperparam in hyperparams.strip().split('\n'):
        try:
            key, min_val, max_val, num_samples, dist, dist_type = hyperparam.strip().split()
            ans[key] = (float(min_val), float(max_val), int(num_samples), dist, dist_type)
        except: ### NOTE: may be fragile
            key, list_ = hyperparam.strip().split(None, 1)
            ans[key] = eval(list_)
    return ans

def evalkwargs(**kwargs):
    '''
    how we parse/evaluate config file specifications into python objects
    essentially follows JSON notation for most things...

    NOTE: we introduce protected options here that cause special behavior. Specifically, "kafka_source" will trigger special parsing logic encapsulated in config2kafka_source
    '''
    for key, val in kwargs.items():
        try:
            val = eval(val) ### NOTE, this may be fragile but should allow a lot of flexibility in how we specify options
        except Exception as e: ### NOTE: this could be fragile...
            pass

        kwargs[key] = val

    return kwargs

def add_missing_kwargs(kwargs, **new_kwargs):
    """
    add kwargs to existing ones if they are not present.
    """
    for key, val in new_kwargs.items():
        if key not in kwargs:
            kwargs[key] = val

    return kwargs

#-------------------------------------------------
### configs for synthetic data

def path2streams(path, segs, verbose=False):
    """
    read in parameters from config and set up the appropriate data structures
    """
    config = path2config(path, verbose=verbose)

    ### iterate through sections (streams) and add their configs to the object
    stream_names = config.sections()

    streams = []
    channel2streams = defaultdict(list)
    for stream in stream_names:
        options = dict(config.items(stream))

        ### iterate through required fields. NOTE: this is a bit fragile and relies on correct matching to MOCKCLASSIFIER_COLUMNS by-hand
        if 'rate' not in options:
            raise KeyError('must specify "rate" for '+stream)
        rate = float(options.pop('rate'))

        if 'frequency distribution' not in options:
            raise KeyError('must specify "frequency distribution" for '+stream)
        freq_distrib = config2feature_distrib(options.pop('frequency distribution'))

        if 'snr distribution' not in options:
            raise KeyError('must specify "snr distribution" for '+stream)
        snr_distrib = config2feature_distrib(options.pop('snr distribution'))

        # instantiate the stream object
        stream = synthetic.SyntheticTriggerStream(stream, segs, rate, freq_distrib, snr_distrib)
        streams.append(stream)

        ### the rest of the options are channels and jitters, so parse them out accordingly
        for channel, jitters in options.items():
            channel2streams[channel].append((stream, config2feature_jitters(jitters)))

    return streams, channel2streams

def config2feature_distrib(string):
    """
    a parsing script for reading off feature distributions
    """
    distrib = string.split()
    args = [float(_) for _ in distrib[1:]] ### FIXME: order of arguments is a bit fragile here and it'll be quite opaque in the config file...
    return synthetic.DistribWrapper(distrib[0], *args)

def config2feature_jitters(string):
    """
    a parsing script for reading off jitter from a string
    """
    ans = {}
    for feature in string.strip().split('\n'):
        if feature:
            key, std = feature.split()
            ans[key] = float(std)
    return ans
